
/*
 * xpa.h
 *
 * SPDX-License-Identifier:  BSD-3-Clause
 *
 * Copyright (C) 2024 brummer <brummer@web.de>
 */

/****************************************************************
  XPa - a C++ wrapper for the portaudio server

  silent the portaudio device probe messages
  connection preference is set to 1.) jackd, 2.) pulse audio, 3.) alsa 

****************************************************************/

#include <portaudio.h>

#if defined(__linux__) || defined(__FreeBSD__) || \
    defined(__NetBSD__) || defined(__OpenBSD__)
#include <pa_jack.h>
#endif

#include <cmath>
#include <vector>
#include <cstdio>
#include <unistd.h>
#include <iostream>
#include <sstream>
#include <string>
#include <stdio.h>

#pragma once

#ifndef XPA_H
#define XPA_H

class XPa {
public:

    XPa(const char* cname){
        #if defined(__linux__) || defined(__FreeBSD__) || \
            defined(__NetBSD__) || defined(__OpenBSD__)
        PaJack_SetClientName (cname);
        #endif
        init();
        SampleRate = 0;
    };

    ~XPa(){Pa_Terminate();};

    // open a audio stream for input/output channels and set the audio process callback
    bool openStream(uint32_t ichannels, uint32_t ochannels, PaStreamCallback *process, void* arg) {
        #if defined(__linux__) || defined(__FreeBSD__) || \
            defined(__NetBSD__) || defined(__OpenBSD__)
        std::vector<Devices> devices;
        int d = Pa_GetDeviceCount();
        const PaDeviceInfo *info;
        for (int i = 0; i<d;i++) {
            info = Pa_GetDeviceInfo(i);
            if ((std::strcmp(info->name, "pulse") ==0) ||
               (std::strcmp(info->name, "default") ==0) ||
               (std::strcmp(info->name, "system") ==0)) {
                Devices dev;
                if (std::strcmp(info->name, "pulse") ==0) dev.order = 2; // pulse audio
                if (std::strcmp(info->name, "default") ==0) dev.order = 3; // alsa
                if (std::strcmp(info->name, "system") ==0) dev.order = 1; // jackd
                dev.index = i;
                dev.SampleRate = SampleRate = info->defaultSampleRate;
                devices.push_back(dev);
            }
            
        }

        std::sort(devices.begin(), devices.end(), 
        [](Devices const &a, Devices const &b) {
            return a.order < b.order; 
        });
        auto it = devices.begin();
        PaStreamParameters inputParameters;
        inputParameters.device = it->index;
        inputParameters.channelCount = ichannels;
        inputParameters.sampleFormat = paFloat32;
        inputParameters.hostApiSpecificStreamInfo = nullptr;

        PaStreamParameters outputParameters;
        outputParameters.device = it->index;
        outputParameters.channelCount = ochannels;
        outputParameters.sampleFormat = paFloat32;
        outputParameters.hostApiSpecificStreamInfo = nullptr;

        err = Pa_OpenStream(&stream, ichannels ? &inputParameters : nullptr, 
                            ochannels ? &outputParameters : nullptr, it->SampleRate,
                            paFramesPerBufferUnspecified, paClipOff, process, arg);

        devices.clear();
        #else
        PaStreamParameters inputParameters;
        inputParameters.device = Pa_GetDefaultInputDevice();
        inputParameters.channelCount = ichannels;
        inputParameters.sampleFormat = paFloat32;
        inputParameters.suggestedLatency = 0.050;
        inputParameters.hostApiSpecificStreamInfo = nullptr;

        PaStreamParameters outputParameters;
        outputParameters.device = Pa_GetDefaultOutputDevice();
        if (outputParameters.device == paNoDevice) return false;
        outputParameters.channelCount = ochannels;
        outputParameters.sampleFormat = paFloat32;
        outputParameters.suggestedLatency = 0.050;
        outputParameters.hostApiSpecificStreamInfo = nullptr;
        SampleRate = 44100;
        err = Pa_OpenStream(&stream, ichannels ? &inputParameters : nullptr, 
                            ochannels ? &outputParameters : nullptr, SampleRate,
                            paFramesPerBufferUnspecified, paClipOff, process, arg);
        #endif

        return err == paNoError ? true : false;
    }

    // start the audio processing
    bool startStream() {
        err = Pa_StartStream(stream);
        return err == paNoError ? true : false;
    }

    // helper function to get a pointer to the PaStream object
    PaStream* getStream() {
        return stream;
    }

    // helper function to get the SampleRate used by the audio sever
    uint32_t getSampleRate() {
        return SampleRate;
    }

    // stop the audio processing
    void stopStream() {
        if (Pa_IsStreamActive(stream)) {
            err = Pa_StopStream(stream);
            if (err != paNoError) {
                std::cerr << "PortAudio error: " << Pa_GetErrorText(err) << std::endl;
            }
            err = Pa_CloseStream(stream);
            if (err != paNoError) {
                std::cerr << "PortAudio error: " << Pa_GetErrorText(err) << std::endl;
            }
        }
    }

private:
    PaStream* stream;
    PaError err;
    uint32_t SampleRate;

    struct Devices {
        int order;
        int index;
        uint32_t SampleRate;
    };

    // initialise the portaudio server,
    // catch the falling device probe messages  
    void init() {
        #if defined(__linux__) || defined(__FreeBSD__) || \
            defined(__NetBSD__) || defined(__OpenBSD__)
        char buffer[1024];
        auto fp = fmemopen(buffer, 1024, "w");
        if ( !fp ) { std::printf("error"); }
        auto old = stderr;
        stderr = fp;
        #endif
        Pa_Initialize();
        #if defined(__linux__) || defined(__FreeBSD__) || \
            defined(__NetBSD__) || defined(__OpenBSD__)
        std::fclose(fp);
        stderr = old;
        #endif
    }

};


#endif
